"""
Consume Daemon main.

"""
from abc import abstractproperty

from microcosm_daemon.api import SleepNow
from microcosm_daemon.daemon import Daemon


class ConsumerDaemon(Daemon):

    def make_arg_parser(self):
        parser = super(ConsumerDaemon, self).make_arg_parser()
        parser.add_argument("--sqs-queue-url")
        return parser

    @abstractproperty
    def schema_mappings(self):
        """
        Define the PubSub message media-type to schema mappings.

        """
        pass

    def handler_mappings(self):
        """
        Define the PubSub message media-type to handler mappings.

        This function is an alternative to configuring `graph.sqs_message_handlers` as
        a graph component; it's useful for very simple mappings (e.g. that don't need access
        to the graph.)

        """
        return {}

    @property
    def defaults(self):
        dct = dict(
            pubsub_message_codecs=dict(
                mappings=self.schema_mappings,
            ),
        )
        if self.handler_mappings:
            dct.update(
                sqs_message_dispatcher=dict(
                    mappings=self.handler_mappings,
                ),
            )
        if self.args.sqs_queue_url:
            dct.update(
                sqs_consumer=dict(
                    sqs_queue_url=self.args.sqs_queue_url,
                ),
            )
        return dct

    @property
    def components(self):
        return super(ConsumerDaemon, self).components + [
            "pubsub_message_codecs",
            "sqs_consumer",
            "sqs_message_dispatcher",
        ]

    def __call__(self, graph):
        """
        Implement daemon by sinking messages from the consumer to a dispatcher function.

        """
        result = graph.sqs_message_dispatcher.handle_batch()
        if not result.message_count:
            raise SleepNow()



"""
Message producer.

"""
from collections import defaultdict

from boto3 import client
from microcosm.api import defaults

from microcosm_pubsub.errors import TopicNotDefinedError


class SNSProducer(object):
    """
    Produces messages to SNS topics.

    """
    def __init__(self, sns_client, sns_topic_arns, pubsub_message_codecs):
        self.sns_client = sns_client
        self.sns_topic_arns = sns_topic_arns
        self.pubsub_message_codecs = pubsub_message_codecs

    def produce(self, media_type, dct=None, **kwargs):
        """
        Produce a message.

        :returns: the message id

        """
        topic_arn = self.choose_topic_arn(media_type)
        content = self.pubsub_message_codecs[media_type].encode(dct, **kwargs)
        result = self.sns_client.publish(
            TopicArn=topic_arn,
            Message=content,
        )
        return result["MessageId"]

    def choose_topic_arn(self, media_type):
        """
        Choose a topic for this type of message.

        """
        try:
            topic_arn = self.sns_topic_arns[media_type]
        except KeyError:
            topic_arn = None

        if topic_arn is None:
            raise TopicNotDefinedError("No topic arn was registered for messages of type: {}".format(
                media_type,
            ))
        return topic_arn


@defaults(
    default=None,
    mappings={},
)
def configure_sns_topic_arns(graph):
    """
    Configure a mapping from message types to topic ARNs.

    """
    if graph.config.sns_topic_arns.default is None:
        sns_topic_arns = dict()
    else:
        sns_topic_arns = defaultdict(lambda: graph.config.sns_topic_arns.default)

    sns_topic_arns.update(graph.config.sns_topic_arns.mappings)

    return sns_topic_arns


def configure_sns_producer(graph):
    """
    Configure an SNS producer.

    """
    if graph.metadata.testing:
        from mock import MagicMock
        sns_client = MagicMock()
    else:
        sns_client = client("sns")

    return SNSProducer(
        sns_client=sns_client,
        sns_topic_arns=graph.sns_topic_arns,
        pubsub_message_codecs=graph.pubsub_message_codecs,
    )



"""
Process batches of messages.

"""
from collections import namedtuple
from logging import getLogger

from microcosm.api import defaults
from microcosm.errors import NotBoundError
from microcosm_logging.decorators import context_logger


DispatchResult = namedtuple("DispatchResult", ["message_count", "error_count", "ignore_count"])


logger = getLogger("pubsub.sink")


class SQSMessageDispatcher(object):
    """
    Dispatch batches of SQSMessages to handler functions.

    """
    def __init__(self, sqs_consumer, sqs_message_handlers, sqs_message_context):
        self.sqs_consumer = sqs_consumer
        self.sqs_message_handlers = sqs_message_handlers
        self.sqs_message_context = sqs_message_context

    def handle_batch(self):
        """
        Send a batch of messages to a function.

        :returns: a `DispatchResult` with metrics around message consumption
        """
        message_count = error_count = ignore_count = 0
        for message in self.sqs_consumer.consume():
            message_count += 1
            try:
                with message:
                    if not self.handle_message(message.media_type, message.content):
                        ignore_count += 1
            except Exception:
                logger.info(
                    "Error handling SQS message.",
                    exc_info=True,
                    extra=self.sqs_message_context(message.content)
                )
                error_count += 1

        return DispatchResult(message_count, error_count, ignore_count)

    def handle_message(self, media_type, message):
        """
        Handle a single message.

        """
        if message is None:
            logger.debug("Skipping message with unparsed type: {}".format(media_type))
            return False
        sqs_message_handler = self.sqs_message_handlers.get(media_type)
        if sqs_message_handler is None:
            logger.debug("Skipping message with unsupported type: {}".format(media_type))
            return False

        handler_with_context = context_logger(
            self.sqs_message_context,
            sqs_message_handler,
            parent=sqs_message_handler,
        )
        return handler_with_context(message)


@defaults(
    mappings=dict(),
)
def configure_sqs_message_dispatcher(graph):
    """
    Configure dispatching of SQS messages.

    Requires a dictionary mapping from the message media type to a handler function; unmapped
    messages will be ignored. This dictionary can either be defined using the `sqs_message_dispatcher.mappings`
    configuration key (which is convenient for simple handlers) or, more likely, via a separate
    `sqs_message_handlers` graph component (which is convenient of the handlers need other component collaborators).
    """
    try:
        sqs_message_handlers = graph.sqs_message_handlers
    except NotBoundError:
        sqs_message_handlers = graph.config.sqs_message_dispatcher.mappings

    try:
        sqs_message_context = graph.sqs_message_context
    except NotBoundError:
        def context(message):
            return dict()
        sqs_message_context = context

    return SQSMessageDispatcher(
        sqs_consumer=graph.sqs_consumer,
        sqs_message_handlers=sqs_message_handlers,
        sqs_message_context=sqs_message_context,
    )



"""
Command line entry point.

"""
from argparse import ArgumentParser
from time import time
from uuid import uuid4

from marshmallow import fields
from microcosm.api import create_object_graph

from microcosm_pubsub.daemon import ConsumerDaemon
from microcosm_pubsub.codecs import PubSubMessageSchema


class SimpleSchema(PubSubMessageSchema):
    """
    A single schema that just sends a text string.

    """
    MEDIA_TYPE = "application/vnd.globality.pubsub.simple"

    message = fields.String(required=True)
    timestamp = fields.Float(required=True)

    def deserialize_media_type(self, obj):
        return SimpleSchema.MEDIA_TYPE


def produce():
    """
    Produce test messages.

    """
    parser = ArgumentParser()
    parser.add_argument("--count", default=1, type=int)
    parser.add_argument("--message")
    parser.add_argument("--message-type", default="test")
    parser.add_argument("--topic-arn", required=True)
    args = parser.parse_args()

    def load_config(metadata):
        return dict(
            pubsub_message_codecs=dict(
                default=SimpleSchema,
            ),
            sns_topic_arns=dict(
                default=args.topic_arn,
            ),
        )

    graph = create_object_graph("example", loader=load_config)
    for _ in range(args.count):
        message_id = graph.sns_producer.produce(
            args.message_type,
            message=args.message or uuid4().hex,
            timestamp=time(),
        )
        print message_id  # noqa


def consume():
    parser = ArgumentParser()
    parser.add_argument("--queue-url", required=True)
    args = parser.parse_args()

    def load_config(metadata):
        return dict(
            pubsub_message_codecs=dict(
                default=SimpleSchema,
            ),
            sqs_consumer=dict(
                sqs_queue_url=args.queue_url,
            ),
        )

    graph = create_object_graph("example", loader=load_config)
    messages = graph.sqs_consumer.consume()

    for message in messages:
        with message:
            print message.content  # noqa


class SimpleConsumerDaemon(ConsumerDaemon):

    @property
    def name(self):
        return "example"

    @property
    def handler_mappings(self):
        return {
            SimpleSchema.MEDIA_TYPE: self.handle_simple,
        }

    @property
    def schema_mappings(self):
        return {
            SimpleSchema.MEDIA_TYPE: SimpleSchema,
        }

    def handle_simple(self, message):
        elapsed = time() - message["timestamp"]
        print "Processed message '{}' after {:.2f}s".format(  # noqa
            message["message"],
            elapsed,
        )
        return True


def main():
    """
    Command line entry point.

    """
    daemon = SimpleConsumerDaemon()
    daemon.run()
